program main
  use precision
  use Matrix_data
  use mm_coo_cuda_csr_convert
  use SP_solvers
  implicit none

    integer                  :: i, j, istat
    integer,parameter        :: iunit1 = 10       ! for Matrix Market data (COO)
    integer,parameter        :: iunit2 = 11       ! for LHS, if available
    character*80             :: mm_file, rhs_file
    integer                  :: extendSymMatrix   ! 1 = symmetrize the pattern (transform from triangular to full)

    integer                  :: gpuonly           ! 0 = both CPU and GPU, 1 = GPU only
    integer                  :: solver            ! 1 = cholesky, 2 = LU, 3 = QR
    integer                  :: reorder           ! Reordering ; 0 = no, 1 = symrcm, 2 = symamd reordering
!   real(fp_kind), parameter :: tol = 0.0d0       ! tolerance to decide if singular or not
    real(fp_kind), parameter :: tol = 1.0d-12     ! tolerance to decide if singular or not
    integer, parameter       :: index_base = 1    ! CSR format -- one based index

! Kind of Solver, a scheme of reordering

    print *, "0 = CPU and GPU, 1 = GPU only?"
    read (5, '(i1)')  gpuonly
    print *, "1 = cholesky, 2 = LU, 3 = QR  ; What solver?"
    read (5, '(i1)')  solver
    if ( .not. (solver == 1 .or. solver == 2 .or. solver == 3) ) solver = 1 

    print *, "0 = No, schemes 1 = symrcm, 2 = symamd ; What kind of reordering?"
    read (5, '(i1)')  reorder

    print '(/1x,a)', "================================================="
    if ( solver == 1) print *, "Sparse matrix linear solver schem : Cholesky"
    if ( solver == 2) print *, "Sparse matrix linear solver schem : LU"
    if ( solver == 3) print *, "Sparse matrix linear solver schem : QR"

    if ( reorder == 1 ) then
      print *, "Reordering scheme of matrix : symrcm"
    else if ( reorder == 2 ) then
      print *, "Reordering scheme of matrix : symamd"
    else
      reorder = 0
      print *, "No reordering of matrix"
    end if
    print '(1x,a/)', "================================================="

! Read matrix Market data file

    extendSymMatrix = 1        !  if =1 .and. symmetric matrix=yes , transform from triangular to full

    print *, "File name for Matrix Market?"
       read (5, '(a)')  mm_file
    !                                mm_file ="lap2D_5pt_n100.mtx"
    print '(/1x,a,a)', "Matrix Market file name = ",mm_file
    open (iunit1, file=mm_file, status='old')
    print *
    call mm_matrix_read(iunit1, extendSymMatrix)

! Read right hand side vectors, if available.

    print *, "File name for LHS vector? if no available, press ENTER and set as RHS=1.0."
    print *
    read (5, '(a)')  rhs_file
    if (rhs_file == "") then
       allocate (rhs(rows))
       rhs= 1.0_fp_kind
    else
      print *, "rhs_file name = ",rhs_file
      open (iunit2, file=rhs_file, status='old')
      call mm_rhs_read(iunit2)
    end if

! Convert COO to CSR sparse matrix
!   INPUT   : Matrix COO data from Matrix Market format via mm_rhs_read routine
!   OUTPUT  : CSR data = csrRowPtrA(m+1), csrColIndA(nz), csrValA(nz) -- global data

    call coo2csr_convert

! CSR sparse matrix data are set. The belows are arguments for cusplverSPs routine.
! Arrays for CUDA cuSPARSE CSR format
!   integer       :: m                 IN    ! rows size for square matrix
!   integer       :: n                 IN    ! columns size for square matrix (n=m)
!   integer       :: nz                IN    ! nonzero element of A
!   integer       :: csrRowPtrA (m+1)  IN    ! the array of offsets corresponding to the start of each row
!   integer       :: csrColIndA (nz)   IN    ! sorted by row and by column within each row
!   real(fp_kind) :: csrValA    (nz)   IN    ! sorted by row and by column within each row
!   real(fp_kind) :: x(m)              INOUT ! initial guess & result
!   real(fp_kind) :: rhs(m)            IN    ! right hand side

! Print CSR Matrix data
!    print *, "csrValA", csrValA
!    print *, "csrRowPtrA",csrRowPtrA
!    print *, "csrColIndA", csrColIndA


! Initilize X
    allocate (x(n))
    x = 0.0_fp_kind      ! or Initial guess X is set

! cuSOLVER SP (Cholesky, LU, QR)

    call cusolverSPs (gpuonly, solver, reorder, m, n, nz, csrRowPtrA, csrColIndA, csrValA, x, rhs, &
                      tol, index_base)

end program main
